from decimal import Decimal
from typing import Union

from django.utils.translation import gettext as _

__all__ = ('encrypt_text', 'decrypt_text', 'encrypt_email_address', 'decrypt_email_address', 'numerize_key', 'NUMBERS',
           'LETTERS_UPPER', 'LETTERS_ONLY', 'ALL_CHARS', 'SYMBOLS', 'LETTERS_ALL', 'LETTERS_ALL_WITH_SPACE',
           'NUMBERS_WITHOUT_ZERO', 'JSON_SAFE_CHARS', 'translate_text', 'translate_email_address', 'translate_iban',
           'translate_number')

# Vigenere like Cipher (Polyalphabetic Substitution Cipher)

NUMBERS_WITHOUT_ZERO = '123456789'
NUMBERS = '1234567890'
LETTERS_ONLY = 'abcdefghijklmnopqrstuvwxyz'
LETTERS_UPPER = LETTERS_ONLY.upper()
LETTERS_ALL = LETTERS_ONLY + LETTERS_UPPER
LETTERS_ALL_WITH_SPACE = LETTERS_ALL + ' '
SYMBOLS = '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~'
# Translators: You can add special characters of your language at the end
ALL_CHARS = _(' !"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~')
# Translators: You can add special characters of your language at the end
JSON_SAFE_CHARS = _(' !#$%&()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[]^_`abcdefghijklmnopqrstuvwxyz{|}~')

# https://en.wikipedia.org/wiki/Email_address#Syntax
RESTRICTED_EMAIL_LOCAL_CHARS = LETTERS_ONLY + NUMBERS + '_-'
EMAIL_LOCAL_CHARS = '!#$%&\'*+-/0123456789=?ABCDEFGHIJKLMNOPQRSTUVWXYZ^_`abcdefghijklmnopqrstuvwxyz{|}~'
DOMAIN_CHARS = LETTERS_ONLY + NUMBERS  # RFC952 + RFC1123


def translate_text(key: str, text: str, encrypt: bool = True, alphabet: str = ALL_CHARS) -> str:
    """
    Translate text based on polyalphabetic substitution cipher based on Vigenere's cipher.

    Notes
        * When the ``translate_text`` encounters a char which is not included in the alphabet the char is skipped and
          left unchanged therefore it is recommended to use on CharFieldAnonymizer the ``transliterate`` option or do
          this yourself on your own anonymizer with method ``unidecode.unidecode``.

    Examples:
        * ``translate_text('LoremIpsum', 'Hello World')`` -> ``'tU_R]IHchZ1'``
        * ``translate_text('LoremIpsum', 'tU_R]IHchZ1', encrypt=False)`` -> ``'Hello World'``
        * ``translate_text('LoremIpsum', 'Hello World', alphabet=LETTERS_ALL)`` -> ``'Hdzcs Waqav'``
        * ``translate_text('LoremIpsum', 'Hdzcs Waqav', encrypt=False, alphabet=LETTERS_ALL)`` -> ``'Hello World'``

    Args:
        key: The encryption key to be used to encrypt or decrypt message
        text: The text to be encrypted or decrypted
        encrypt: If ``True`` the function encrypts the text. If ``False`` the function decrypts the text.
        alphabet: The "alphabet" to be used for encryption thanks to this you can encrypt for example only numbers.

    Returns:
        Encrypted or decrypted text

    """
    translated = []

    key_index = 0
    for char in text:
        num = alphabet.find(char)
        if num != -1:
            num += (1 if encrypt else -1) * alphabet.find(key[key_index])
            num %= len(alphabet)

            translated.append(alphabet[num])

            key_index += 1
            if key_index == len(key):
                key_index = 0
        else:
            translated.append(char)

    return "".join(translated)


def encrypt_text(key: str, text: str, alphabet: str = ALL_CHARS) -> str:
    """
    Encrypts text based on polyalphabetic substitution cipher based on Vigenere's cipher.

    Notes
        * When the ``encrypt_text`` encounters a char which is not included in the alphabet the char is skipped and
          left unchanged therefore it is recommended to use on CharFieldAnonymizer the ``transliterate`` option or
          do this yourself on your own anonymizer with method `unidecode.unidecode`.

    Examples:
        * ``encrypt_text('LoremIpsum', 'Hello World')`` -> ``'tU_R]IHchZ1'``
        * ``encrypt_text('LoremIpsum', 'Hello World', alphabet=LETTERS_ALL)`` -> ``'Hdzcs Waqav'``

    See Also:
        * ``gdpr.encryption.translate_text``
        * ``gdpr.encryption.decrypt_text``

    Args:
        key: The encryption key
        text: The plaintext to be encrypted
        alphabet: The "alphabet" to be used for encryption thanks to this you can encrypt for example only numbers.

    Returns:
        Ciphertext - Encrypted text

    """
    return translate_text(key, text, True, alphabet)


def decrypt_text(key: str, text: str, alphabet: str = ALL_CHARS) -> str:
    """
    Decrypts text based on polyalphabetic substitution cipher based on Vigenere's cipher.

    Notes
        * When the ``decrypt_text`` encounters a char which is not included in the alphabet the char is skipped and
          left unchanged therefore it is recommended to use on CharFieldAnonymizer the ``transliterate`` option or do
          this yourself on your own anonymizer with method `unidecode.unidecode`.

    Examples:
        * ``decrypt_text('LoremIpsum', 'tU_R]IHchZ1')`` -> ``'Hello World'``
        * ``decrypt_text('LoremIpsum', 'Hdzcs Waqav`', alphabet=LETTERS_ALL)` ->``'Hello World'``

    See Also:
            * ``gdpr.encryption.translate_text``
            * ``gdpr.encryption.encrypt_text``

    Args:
        key: The encryption key
        text: The ciphertext to be decrypted
        alphabet: The "alphabet" to be used for encryption thanks to this you can encrypt for example only numbers.

    Returns:
        Plaintext - Decrypted text

    """
    return translate_text(key, text, False, alphabet)


def translate_email_address(key: str, email: str, encrypt: bool = True, restricted_mode: bool = True):
    """
    Translate email address based on polyalphabetic substitution cipher based on Vigenere's cipher.

    Examples:
        * ``translate_email_address('LoremIpsum', 'foo@bar.com')`` -> ``'-_b@ao9.com'``
        * ``translate_email_address('LoremIpsum', '-_b@ao9.com', encrypt=False)`` -> ``'foo@bar.com'``

    Notes:
        * The first part of the email address in non restricted mode is translated based on RFC5322, RFC5321, RFC3696
          and domain name on RFC952 + RFC1123
        * The translations skips dots (``.``) as they cannot be at the begging or end.
        * The TLD (Top level domain) is not translated to serve statistical purposes.

    Warnings:
        * The translation method *does not support* comments or IP address in the domain part of the email e.g.: Formats
          ``jsmith@[192.168.2.1]`` or ``jsmith@[IPv6:2001:db8::1]`` and
          ``john.smith@(comment)example.com`` or ``john.smith@example.com(comment)`` are not supported.
        * The input is not being checked for correct format.

    See Also:
        * ``gdpr.encryption.translate_text``
        * https://en.wikipedia.org/wiki/Email_address#Local-part

    Args:
        key: The encryption key
        email: The email address to be encrypted or decrypted
        encrypt: If ``True`` the function encrypts the email. If ``False`` the function decrypts the email.
        restricted_mode: Some email clients such as Windows Live Hotmail does not permit special characters despite
                         their standardized used. If this option is ``True`` encryption will use only alphanumeric
                         chars, underscore(``_``) and hyphen (``-``).

    Returns:
        Encrypted or decrypted email address

    """
    local, domain_tld = email.split("@")
    domain, tld = domain_tld.rsplit(".", 1)
    if not restricted_mode:
        return (f'{translate_text(key, local, encrypt, EMAIL_LOCAL_CHARS)}@'
                f'{translate_text(key, domain, encrypt, DOMAIN_CHARS)}.{tld}')
    else:
        return (f'{translate_text(key, local, encrypt, RESTRICTED_EMAIL_LOCAL_CHARS)}@'
                f'{translate_text(key, domain, encrypt, DOMAIN_CHARS)}.{tld}')


def encrypt_email_address(key: str, email: str, restricted_mode: bool = True):
    """
    Encrypts email address based on polyalphabetic substitution cipher based on Vigenere's cipher.

    Examples:
        * ``encrypt_email_address('LoremIpsum', 'foo@bar.com')`` -> ``'-_b@ao9.com'``

    Notes:
        * The first part of the email address in non restricted mode is translated based on RFC5322, RFC5321, RFC3696
          and domain name on RFC952 + RFC1123
        * The translations skips dots (``.``) as they cannot be at the begging or end.
        * The TLD (Top level domain) is not translated to serve statistical purposes.

    Warnings:
        * The translation method *does not support* comments or IP address in the domain part of the email e.g.: Formats
          ``jsmith@[192.168.2.1]`` or ``jsmith@[IPv6:2001:db8::1]`` and
          ``john.smith@(comment)example.com`` or ``john.smith@example.com(comment)`` are not supported.
        * The input is not being checked for correct format.

    See Also:
        * ``gdpr.encryption.translate_text``
        * ``gdpr.encryption.translate_email_address``
        * ``gdpr.encryption.decrypt_email_address``
        * https://en.wikipedia.org/wiki/Email_address#Local-part

    Args:
        key: The encryption key
        email: The email address to be encrypted
        restricted_mode: Some email clients such as Windows Live Hotmail does not permit special characters despite
                         their standardized used. If this option is `True` encryption will use only alphanumeric chars,
                         underscore(``_``) and hyphen (``-``).

    Returns:
        Encrypted email address

    """
    return translate_email_address(key, email, True, restricted_mode)


def decrypt_email_address(key: str, email: str, restricted_mode: bool = True):
    """
    Decrypts email address based on polyalphabetic substitution cipher based on Vigenere's cipher.

    Examples:
        * ``decrypt_email_address('LoremIpsum', '-_b@ao9.com')`` -> ``'foo@bar.com'``

    Notes:
        * The first part of the email address in non restricted mode is translated based on RFC5322, RFC5321, RFC3696
          and domain name on RFC952 + RFC1123
        * The translations skips dots (``.``) as they cannot be at the begging or end.
        * The TLD (Top level domain) is not translated to serve statistical purposes.

    Warnings:
        * The translation method *does not support* comments or IP address in the domain part of the email e.g.: Formats
          ``jsmith@[192.168.2.1]`` or ``jsmith@[IPv6:2001:db8::1]`` and
          ``john.smith@(comment)example.com`` or ``john.smith@example.com(comment)`` are not supported.
        * The input is not being checked for correct format.

    See Also:
        * ``gdpr.encryption.translate_text``
        * ``gdpr.encryption.translate_email_address``
        * ``gdpr.encryption.encrypt_email_address``
        * https://en.wikipedia.org/wiki/Email_address#Local-part

    Args:
        key: The encryption key
        email: The email address to be encrypted or decrypted
        encrypt: If ``True`` the function encrypts the email. If ``False`` the function decrypts the email.
        restricted_mode: Some email clients such as Windows Live Hotmail does not permit special characters despite
                         their standardized used. If this option is `True` encryption will use only alphanumeric chars,
                         underscore(``_``) and hyphen (``-``).

    Returns:
        Decrypted email address

    """
    return translate_email_address(key, email, False, restricted_mode)


def numerize_key(key: str) -> int:
    """
    Blackbox function for generating some big number from str.

    Examples:
        * ``numerize_key('LoremIpsum')`` -> ``7406395466880``
        * ``numerize_key('LoremIpsumDolorSitAmet')`` -> ``69127628581563401652806784``

    Notes:
        In case you are really wondering how this function works. The algorithm is following:
        1. Each letter convert to ascii code number
        2. Multiply each number by ``42``
        3. we multiple each number by ten to the power of it's position (first position is 0)
        5. We sum all numbers

    Args:
        key: String encryption key

    Returns:
        Big whole number to be used as encryption key

    """
    return sum([ord(key[i]) * 42 ** (i + 1) for i in range(len(key))])


def translate_type_match(key: str, text: str, encrypt: bool = True, numbers: str = NUMBERS,
                         alphabet: str = LETTERS_ALL) -> str:
    """
    Translate text while retaining type (char, number) based on polyalphabetic substitution cipher.

    Examples:
        * ``translate_type_match('LoremIpusm', 'CZ20180612')`` -> ``'BY50901250'``
        * ``translate_type_match('LoremIpusm', 'BY50901250', encrypt=False)`` -> ``'CZ20180612'``

    See Also:
        * ``gdpr.encryption.translate_text``

    Args:
        key: The encryption key
        text: The text to be encrypted or decrypted
        encrypt: If ``True`` the function encrypts the text. If ``False`` the function decrypts the text.
        numbers: The "alphabet" used to translate numbers.
        alphabet: The "alphabet" used to translate letters.

    Returns:
        Encrypted or decrypted text

    """
    translated = []

    key_index = 0
    for char in text:
        is_symbol_number = char in numbers
        alphabet = numbers if is_symbol_number else alphabet
        actual_key = key if not is_symbol_number else str(ord(key[key_index]))[-1]

        num = alphabet.find(char)
        if num != -1:
            num += (1 if encrypt else -1) * alphabet.find(actual_key)
            num %= len(alphabet)

            translated.append(alphabet[num])

            key_index += 1
            if key_index == len(key):
                key_index = 0
        else:
            translated.append(char)

    return "".join(translated)


def translate_iban(key: str, iban: str, encrypt: bool = True) -> str:
    """
    Translate IBAN using ``translate_type_match`` function.

    Warnings:
        * This method only encrypts the numbers and letters, therefore output of this function will have IBAN format
          but will not pass any additional validations. If you need this used localised smart Anonymizer as
          IBAN validation is country specific.

    See Also:
        * ``gdpr.encryption.translate_text``
        * ``gdpr.encryption.translate_type_match``

    Examples:
        * ``translate_iban('LoremIpsum', 'CZ65 0800 0000 1920 0014 5399')`` - > ``'CZ15 3882 1468 6950 8228 1149'``
        * ``translate_iban('LoremIpsum', 'CZ15 3882 1468 6950 8228 1149', encrypt=False)`` ->
          ``'CZ65 0800 0000 1920 0014 5399'``
        * ``translate_iban('LoremIpsum', 'GB29NWBK60161331926819')`` -> ``'GB79MVAJ74746361747277'``
        * ``translate_iban('LoremIpsum', 'GB79MVAJ74746361747277', encrypt=False)`` -> ``'GB29NWBK60161331926819'``

    Args:
        key: The encryption key
        iban: The IBAN to be encrypted or decrypted
        encrypt: If ``True`` the function encrypts the text. If ``False`` the function decrypts the text.

    Returns:
        Encrypted or decrypted IBAN

    """
    return iban[:2].upper() + translate_type_match(key, iban[2:].upper(), encrypt, alphabet=LETTERS_UPPER)


def translate_number(key: str, number: Union[Decimal, int], encrypt: bool = True):
    """
    Takes number converts it to string and translate it using ``translate_text`` with num alphabet.

    Args:
        key: The encryption key
        number: The number Decimal or int to be translated
        encrypt: If ``True`` the function encrypts the text. If ``False`` the function decrypts the text.

    Returns:
        Encrypted number

    """
    original_type = type(number)
    str_number = str(number)
    if "-" in str_number:
        if "." in str_number:
            return original_type(
                str_number[0] + translate_text(
                    key, str_number[1], encrypt=encrypt, alphabet=NUMBERS_WITHOUT_ZERO
                ) + translate_text(
                    key, str_number[2:-1], encrypt=encrypt, alphabet=NUMBERS
                ) + translate_text(key, str_number[-1], encrypt=encrypt, alphabet=NUMBERS_WITHOUT_ZERO)
            )
        else:
            return original_type(
                str_number[0] + translate_text(
                    key, str_number[1], encrypt=encrypt, alphabet=NUMBERS_WITHOUT_ZERO
                ) + translate_text(key, str_number[2:], encrypt=encrypt, alphabet=NUMBERS)
            )
    else:
        if "." in str_number:
            return original_type(
                translate_text(
                    key, str_number[0], encrypt=encrypt, alphabet=NUMBERS_WITHOUT_ZERO
                ) + translate_text(
                    key, str_number[1:-1], encrypt=encrypt, alphabet=NUMBERS
                ) + translate_text(key, str_number[-1], encrypt=encrypt, alphabet=NUMBERS_WITHOUT_ZERO)
            )
        else:
            return original_type(
                translate_text(
                    key, str_number[0], encrypt=encrypt, alphabet=NUMBERS_WITHOUT_ZERO
                ) + translate_text(key, str_number[1:], encrypt=encrypt, alphabet=NUMBERS)
            )



# Generated by Django 2.1.5 on 2019-02-08 09:38

from django.db import migrations, models
import django.db.models.deletion
import gdpr.mixins


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('tests', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Note',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('note', models.TextField()),
                ('object_id', models.PositiveIntegerField()),
                ('content_type',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
            options={
                'abstract': False,
            },
            bases=(gdpr.mixins.AnonymizationModelMixin, models.Model),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.8 on 2019-02-25 16:27
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('gdpr', '0004'),
    ]

    operations = [
        migrations.AlterField(
            model_name='anonymizeddata',
            name='content_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='contenttypes.ContentType',
                                    verbose_name='related object content type'),
        ),
        migrations.AlterField(
            model_name='anonymizeddata',
            name='expired_reason',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL,
                                    to='gdpr.LegalReason', verbose_name='expired reason'),
        ),
        migrations.AlterField(
            model_name='legalreason',
            name='expires_at',
            field=models.DateTimeField(blank=True, db_index=True, null=True, verbose_name='expires at'),
        ),
        migrations.AlterField(
            model_name='legalreason',
            name='source_object_content_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='contenttypes.ContentType',
                                    verbose_name='source object content type'),
        ),
        migrations.AlterField(
            model_name='legalreason',
            name='source_object_id',
            field=models.TextField(verbose_name='source object ID'),
        ),
        migrations.AlterField(
            model_name='legalreasonrelatedobject',
            name='object_content_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='contenttypes.ContentType',
                                    verbose_name='related object content type'),
        ),
        migrations.AlterField(
            model_name='legalreasonrelatedobject',
            name='object_id',
            field=models.TextField(verbose_name='related object ID'),
        ),
        migrations.AlterUniqueTogether(
            name='anonymizeddata',
            unique_together=set([('content_type', 'object_id', 'field')]),
        ),
        migrations.AlterUniqueTogether(
            name='legalreason',
            unique_together=set([('purpose_slug', 'source_object_content_type', 'source_object_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='legalreasonrelatedobject',
            unique_together=set([('legal_reason', 'object_content_type', 'object_id')]),
        ),
    ]



# Generated by Django 2.1.5 on 2019-01-07 16:35

import django.db.models.deletion
from django.db import migrations, models

import gdpr.mixins


class Migration(migrations.Migration):
    initial = True

    dependencies = []  # type: ignore

    operations = [
        migrations.CreateModel(
            name='Account',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('number', models.CharField(blank=True, max_length=256, null=True)),
                ('owner', models.CharField(blank=True, max_length=256, null=True)),
            ],
            options={
                'abstract': False,
            },
            bases=(gdpr.mixins.AnonymizationModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Address',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('street', models.CharField(blank=True, max_length=256, null=True)),
                ('house_number', models.CharField(blank=True, max_length=20, null=True)),
                ('city', models.CharField(blank=True, max_length=256, null=True)),
                ('post_code', models.CharField(blank=True, max_length=6, null=True)),
            ],
            options={
                'abstract': False,
            },
            bases=(gdpr.mixins.AnonymizationModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ContactForm',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('email', models.EmailField(max_length=254)),
                ('full_name', models.CharField(max_length=256)),
            ],
            options={
                'abstract': False,
            },
            bases=(gdpr.mixins.AnonymizationModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Customer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('first_name', models.CharField(max_length=256)),
                ('last_name', models.CharField(max_length=256)),
                ('primary_email_address', models.EmailField(blank=True, max_length=254, null=True)),
                ('full_name', models.CharField(blank=True, max_length=256, null=True)),
                ('birth_date', models.DateField(blank=True, null=True)),
                ('personal_id', models.CharField(blank=True, max_length=10, null=True)),
                ('phone_number', models.CharField(blank=True, max_length=9, null=True)),
                ('fb_id', models.CharField(blank=True, max_length=256, null=True)),
                ('last_login_ip', models.GenericIPAddressField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
            bases=(gdpr.mixins.AnonymizationModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Email',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('email', models.EmailField(blank=True, max_length=254, null=True)),
                ('customer', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='emails',
                                               to='tests.Customer')),
            ],
            options={
                'abstract': False,
            },
            bases=(gdpr.mixins.AnonymizationModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Payment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.DecimalField(blank=True, decimal_places=2, max_digits=10, null=True)),
                ('date', models.DateField(auto_now_add=True)),
                ('account', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='payments',
                                              to='tests.Account')),
            ],
            options={
                'abstract': False,
            },
            bases=(gdpr.mixins.AnonymizationModelMixin, models.Model),
        ),
        migrations.AddField(
            model_name='address',
            name='customer',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='addresses',
                                    to='tests.Customer'),
        ),
        migrations.AddField(
            model_name='account',
            name='customer',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='accounts',
                                    to='tests.Customer'),
        ),
    ]



from dateutil.relativedelta import relativedelta

from django.test import TestCase

from freezegun import freeze_time
from gdpr.models import LegalReason
from germanium.tools import assert_equal, assert_false, assert_not_equal, assert_true, test_call_command
from tests.models import Customer
from tests.purposes import FIRST_AND_LAST_NAME_SLUG, MARKETING_SLUG
from tests.tests.data import CUSTOMER__FIRST_NAME, CUSTOMER__KWARGS
from tests.tests.utils import AnonymizedDataMixin


class TestDeactivateExpiredReasons(AnonymizedDataMixin, TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.customer: Customer = Customer(**CUSTOMER__KWARGS)
        cls.customer.save()

    def test_command_should_expire_active_legal_reson(self):
        legal_reason = LegalReason.objects.create_consent(MARKETING_SLUG, self.customer)
        legal_reason.save()

        # consent has not expired
        test_call_command('deactivate_expired_reasons')
        assert_true(LegalReason.objects.exists_valid_consent(MARKETING_SLUG, self.customer))

        # consent has expired
        with freeze_time(legal_reason.expires_at + relativedelta(seconds=1)):
            test_call_command('deactivate_expired_reasons')
        assert_false(LegalReason.objects.exists_valid_consent(MARKETING_SLUG, self.customer))

    def test_command_should_not_expire_deactivated_legal_reasons(self):
        legal_reason = LegalReason.objects.create_consent(MARKETING_SLUG, self.customer)
        legal_reason.save()
        legal_reason.deactivate()

        # consent has not expired
        test_call_command('deactivate_expired_reasons')
        assert_true(LegalReason.objects.exists_deactivated_consent(MARKETING_SLUG, self.customer))

        # consent has expired
        with freeze_time(legal_reason.expires_at + relativedelta(seconds=1)):
            test_call_command('deactivate_expired_reasons')
        assert_true(LegalReason.objects.exists_deactivated_consent(MARKETING_SLUG, self.customer))

    def test_command_should_not_touch_already_expired_legal_reasons(self):
        legal_reason = LegalReason.objects.create_consent(MARKETING_SLUG, self.customer)
        legal_reason.save()
        legal_reason.expire()
        original_expiration_time = legal_reason.changed_at

        with freeze_time(original_expiration_time + relativedelta(seconds=1)):
            test_call_command('deactivate_expired_reasons')
        assert_equal(original_expiration_time, legal_reason.refresh_from_db().changed_at)

    def test_command_should_not_anonymize_customer_if_expiring_reason_is_not_related_to_customer_data(self):
        marketing_legal_reason = LegalReason.objects.create_consent(MARKETING_SLUG, self.customer)
        marketing_legal_reason.save()
        LegalReason.objects.create_consent(FIRST_AND_LAST_NAME_SLUG, self.customer).save()

        # none of the consents have expired
        test_call_command('deactivate_expired_reasons')
        assert_true(LegalReason.objects.exists_valid_consent(MARKETING_SLUG, self.customer))
        assert_true(LegalReason.objects.exists_valid_consent(FIRST_AND_LAST_NAME_SLUG, self.customer))

        # marketing consent has expired
        with freeze_time(marketing_legal_reason.expires_at + relativedelta(seconds=1)):
            test_call_command('deactivate_expired_reasons')
        assert_false(LegalReason.objects.exists_valid_consent(MARKETING_SLUG, self.customer))
        assert_true(LegalReason.objects.exists_valid_consent(FIRST_AND_LAST_NAME_SLUG, self.customer))
        assert_equal(self.customer.first_name, CUSTOMER__FIRST_NAME)
        self.assertAnonymizedDataNotExists(self.customer, 'first_name')
        assert_not_equal(self.customer._anonymize_obj(), CUSTOMER__FIRST_NAME)
        self.assertAnonymizedDataExists(self.customer, 'first_name')



from datetime import date
from decimal import Decimal


CUSTOMER__FIRST_NAME = "John"
CUSTOMER__LAST_NAME = "Smith"
CUSTOMER__EMAIL = "smth@u.plus"
CUSTOMER__EMAIL2 = "foo@u.plus"
CUSTOMER__EMAIL3 = "bar@u.plus"
CUSTOMER__BIRTH_DATE = date(2000, 1, 1)
CUSTOMER__PERSONAL_ID = "0001010004"
CUSTOMER__PHONE_NUMBER = "605222111"
CUSTOMER__FACEBOOK_ID = "4"
CUSTOMER__IP = "127.0.0.1"

CUSTOMER__KWARGS = {
    "first_name": CUSTOMER__FIRST_NAME,
    "last_name": CUSTOMER__LAST_NAME,
    "primary_email_address": CUSTOMER__EMAIL,
    "birth_date": CUSTOMER__BIRTH_DATE,
    "personal_id": CUSTOMER__PERSONAL_ID,
    "phone_number": CUSTOMER__PHONE_NUMBER,
    "facebook_id": CUSTOMER__FACEBOOK_ID,
    "last_login_ip": CUSTOMER__IP
}

ADDRESS__STREET = "Downing Street"
ADDRESS__HOUSE_NUMBER = "10"
ADDRESS__CITY = "London"
ADDRESS__POST_CODE = "SW1A 2AB"

ACCOUNT__NUMBER = "19-2000145399/0800"
ACCOUNT__IBAN = "CZ65 0800 0000 1920 0014 5399"
ACCOUNT__SWIFT = "GIBACZPX"
ACCOUNT__OWNER = "John Smith"

ACCOUNT__NUMBER2 = "2501277007/2010"
ACCOUNT__IBAN2 = "CZ54 2010 0000 0025 0127 7007"
ACCOUNT__SWIFT2 = "FIOBCZPP"
ACCOUNT__OWNER2 = "James Bond"

PAYMENT__VALUE = Decimal("10.1")



import warnings

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model, QuerySet
from django.db.utils import Error

from gdpr.models import AnonymizedData, LegalReason, LegalReasonRelatedObject


class AnonymizationModelMixin:

    @property
    def content_type(self) -> ContentType:
        """Get model ContentType"""
        return ContentType.objects.get_for_model(self.__class__)

    def _anonymize_obj(self, *args, **kwargs):
        from gdpr.loading import anonymizer_register
        if self.__class__ in anonymizer_register:
            anonymizer_register[self.__class__]().anonymize_obj(self, *args, **kwargs)
        else:
            raise ImproperlyConfigured('%s does not have registered anonymizer.' % self.__class__)

    def _deanonymize_obj(self, *args, **kwargs):
        from gdpr.loading import anonymizer_register
        if self.__class__ in anonymizer_register:
            anonymizer_register[self.__class__]().deanonymize_obj(self, *args, **kwargs)
        else:
            raise ImproperlyConfigured('%s does not have registered anonymizer.' % self.__class__)

    def get_consents(self) -> QuerySet:
        return LegalReason.objects.filter_source_instance(self)

    def create_consent(self, purpose_slug: str, *args, **kwargs) -> LegalReason:
        return LegalReason.objects.create_consent(purpose_slug, self, *args, **kwargs)

    def deactivate_consent(self, purpose_slug: str):
        LegalReason.objects.deactivate_consent(purpose_slug, self)

    def delete(self, *args, **kwargs):
        """Cleanup anonymization metadata"""
        obj_id = str(self.pk)
        super().delete(*args, **kwargs)
        try:
            AnonymizedData.objects.filter(object_id=obj_id, content_type=self.content_type).delete()
        except Error as e:
            # Better to just have some leftovers then to fail
            warnings.warn(f'An exception {str(e)} occurred during cleanup of {str(self)}')
        try:
            LegalReasonRelatedObject.objects.filter(object_id=obj_id, object_content_type=self.content_type).delete()
        except Error as e:
            # Better to just have some leftovers then to fail
            warnings.warn(f'An exception {str(e)} occurred during cleanup of {str(self)}')


class AnonymizationModel(AnonymizationModelMixin, Model):
    class Meta:
        abstract = True



from django.test import TestCase

from germanium.tools import assert_true


class EmptyTests(TestCase):
    def test_this_test_is_run(self):
        """Test that the test suite is running this tests."""
        assert_true(True)

    def test_GDPR_app_is_reachable(self):
        """Test that GDPR app is reachable."""
        from gdpr.version import get_version
        get_version()
        assert_true(True)



import math

import pyprind
from django.core.management.base import BaseCommand
from utils import chunked_iterator, chunked_queryset_iterator
from utils.commands import ProgressBarStream

from gdpr.anonymizers import DeleteModelAnonymizer
from gdpr.loading import anonymizer_register


class Command(BaseCommand):
    help = 'Anonymize database data according to defined anonymizers in applications.'

    def add_arguments(self, parser):
        parser.add_argument('--models', type=str, action='store', dest='models',
                            help='name of the anonymized models ("app_name.model_name") separated by a comma.')

    def _anonymize_by_qs(self, obj_anonymizer, qs):
        bar = pyprind.ProgBar(
            max(math.ceil(qs.count() // obj_anonymizer.chunk_size), 1),
            title='Anonymize model {}'.format(self._get_full_model_name(qs.model)),
            stream=ProgressBarStream(self.stdout)
        )
        for batch_qs in chunked_queryset_iterator(qs, obj_anonymizer.chunk_size, delete_qs=isinstance(
                obj_anonymizer, DeleteModelAnonymizer)):
            obj_anonymizer().anonymize_qs(batch_qs)
            bar.update()

    def _anonymize_by_obj(self, obj_anonymizer, qs):
        bar = pyprind.ProgBar(
            qs.count(),
            title='Anonymize model {}'.format(self._get_full_model_name(qs.model)),
            stream=ProgressBarStream(self.stdout)
        )
        for obj in chunked_iterator(qs, obj_anonymizer.chunk_size):
            obj_anonymizer().anonymize_obj(obj)
            bar.update()

    def _anonymize(self, obj_anonymizer, model):
        qs = model.objects.all()
        if obj_anonymizer.can_anonymize_qs:
            self._anonymize_by_qs(obj_anonymizer, qs)
        else:
            self._anonymize_by_obj(obj_anonymizer, qs)

    def _get_full_model_name(self, model):
        return '{}.{}'.format(model._meta.app_label, model._meta.model_name)

    def handle(self, models, *args, **options):
        models = {v.strip().lower() for v in models.split(',')} if models else None
        for obj_anonymizer in list(anonymizer_register()):
            model = obj_anonymizer.Meta.model
            if not models or self._get_full_model_name(model) in models:
                self._anonymize(obj_anonymizer, model)
        self.stdout.write('Data was anonymized')



from django.test import TestCase

from gdpr.ipcypher import decrypt_ip, decrypt_ipv4, decrypt_ipv6, derive_key, encrypt_ip, encrypt_ipv4, encrypt_ipv6
from germanium.tools import assert_equal


KEY_EMPTY_CLEAR = ''
KEY_EMPTY = bytearray.fromhex('bb8dcd7be9a6f43b3304c640d7d7103c')

KEY_PI_CLEAR = '3.141592653589793'
KEY_PI = bytearray.fromhex('3705bd6c0e26a1a839898f1fa016a374')

KEY_CRYPTO_CLEAR = 'crypto is not a coin'
KEY_CRYPTO = bytearray.fromhex('06c4bad23a38b9e0ad9d0590b0a3d93a')

KEY = 'crypto is not a coin'


class TestIPCrypt(TestCase):
    def test_derivation_empty(self):
        key = derive_key(KEY_EMPTY_CLEAR)

        assert_equal(KEY_EMPTY, key)

    def test_derivation_pi(self):
        key = derive_key(KEY_PI_CLEAR)

        assert_equal(KEY_PI, key)

    def test_derivation_crypto(self):
        key = derive_key(KEY_CRYPTO_CLEAR)

        assert_equal(KEY_CRYPTO, key)

    def test_ipv4_key_1(self):
        clear_ip = '198.41.0.4'
        encrypted_ip = encrypt_ipv4(KEY, clear_ip)
        decrypted_ip = decrypt_ipv4(KEY, encrypted_ip)

        assert_equal(encrypted_ip, '139.111.117.167')
        assert_equal(decrypted_ip, clear_ip)

    def test_ipv4_key_2(self):
        clear_ip = '130.161.180.1'
        encrypted_ip = encrypt_ipv4(KEY, clear_ip)
        decrypted_ip = decrypt_ipv4(KEY, encrypted_ip)

        assert_equal(encrypted_ip, '66.235.221.231')
        assert_equal(decrypted_ip, clear_ip)

    def test_ipv4_key_3(self):
        clear_ip = '0.0.0.0'
        encrypted_ip = encrypt_ipv4(KEY, clear_ip)
        decrypted_ip = decrypt_ipv4(KEY, encrypted_ip)

        assert_equal(encrypted_ip, '203.253.152.187')
        assert_equal(decrypted_ip, clear_ip)

    def test_ipv6_key_1(self):
        clear_ip = '::1'
        encrypted_ip = encrypt_ipv6(KEY, clear_ip)
        decrypted_ip = decrypt_ipv6(KEY, encrypted_ip)

        assert_equal(encrypted_ip, 'a551:9cb0:c9b:f6e1:6112:58a:af29:3a6c')
        assert_equal(decrypted_ip, clear_ip)

    def test_ipv6_key_2(self):
        clear_ip = '2001:503:ba3e::2:30'
        encrypted_ip = encrypt_ipv6(KEY, clear_ip)
        decrypted_ip = decrypt_ipv6(KEY, encrypted_ip)

        assert_equal(encrypted_ip, '6e60:2674:2fac:d383:f9d5:dcfe:fc53:328e')
        assert_equal(decrypted_ip, clear_ip)

    def test_ipv6_key_3(self):
        clear_ip = '2001:db8::'
        encrypted_ip = encrypt_ipv6(KEY, clear_ip)
        decrypted_ip = decrypt_ipv6(KEY, encrypted_ip)

        assert_equal(encrypted_ip, 'a8f5:16c8:e2ea:23b9:748d:67a2:4107:9d2e')
        assert_equal(decrypted_ip, clear_ip)

    def test_general_function_ipv4(self):
        clear_ip = '198.41.0.4'
        encrypted_ip = encrypt_ip(KEY, clear_ip)
        decrypted_ip = decrypt_ip(KEY, encrypted_ip)

        assert_equal(encrypted_ip, '139.111.117.167')
        assert_equal(decrypted_ip, clear_ip)

    def test_general_function_ipv6(self):
        clear_ip = '::1'
        encrypted_ip = encrypt_ip(KEY, clear_ip)
        decrypted_ip = decrypt_ip(KEY, encrypted_ip)

        assert_equal(encrypted_ip, 'a551:9cb0:c9b:f6e1:6112:58a:af29:3a6c')
        assert_equal(decrypted_ip, clear_ip)



import os

from gdpr.version import get_version
from setuptools import find_packages, setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='django-GDPR',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    version=get_version(),
    description='Library for GDPR implementation',
    author='Druids',
    author_email='matllubos@gmail.com',
    url='https://github.com/druids/django-GDPR',
    license='MIT',
    package_dir={'gdpr': 'gdpr'},
    include_package_data=True,
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Django',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'django>=2.2',
        'django-chamber>=0.6.6',
        'tqdm>=4.28.1',
        'pyaes>=1.6.1',
        'unidecode',
        'django-choice-enumfields>=1.0.5',
    ],
    zip_safe=False,
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.16 on 2018-12-19 15:17
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('gdpr', '0003'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='legalreason',
            unique_together=set([('purpose_slug', 'source_object_content_type', 'source_object_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='legalreasonrelatedobject',
            unique_together=set([('legal_reason', 'object_content_type', 'object_id')]),
        ),
    ]



# Generated by Django 3.1 on 2021-04-18 10:56

from django.db import migrations
import enumfields.fields
import gdpr.enums


class Migration(migrations.Migration):

    dependencies = [
        ('gdpr', '0010_migration'),
    ]

    operations = [
        migrations.AlterField(
            model_name='legalreason',
            name='state',
            field=enumfields.fields.NumEnumField(
                db_index=True, default=1, enum=gdpr.enums.LegalReasonState, verbose_name='state'
            ),
        ),
    ]



from django.core.management.base import BaseCommand

from gdpr.models import LegalReason

from chamber.utils.tqdm import tqdm

from django.db import transaction


class Command(BaseCommand):

    def handle(self, *args, **options):
        self.stdout.write('Anonymize expired data of expired legal reasons')
        legal_reason_to_expire_qs = LegalReason.objects.filter_active_and_expired()
        for legal_reason in tqdm(legal_reason_to_expire_qs.iterator(),
                                 total=legal_reason_to_expire_qs.count(),
                                 file=self.stdout):
            legal_reason.expire()



from .fields import (
    CharFieldAnonymizer, DateFieldAnonymizer, DateTimeFieldAnonymizer, DecimalFieldAnonymizer, EmailFieldAnonymizer,
    FunctionFieldAnonymizer, IBANFieldAnonymizer, IPAddressFieldAnonymizer, SiteIDUsernameFieldAnonymizer,
    StaticValueFieldAnonymizer, DeleteFileFieldAnonymizer, IntegerFieldAnonymizer, ReplaceFileFieldAnonymizer)
from .generic_relation import GenericRelationAnonymizer, ReverseGenericRelationAnonymizer
from .hash_fields import HashTextFieldAnonymizer, MD5TextFieldAnonymizer, SHA256TextFieldAnonymizer
from .model_anonymizers import DeleteModelAnonymizer, ModelAnonymizer

__all__ = (
    'ModelAnonymizer', 'DeleteModelAnonymizer', 'FunctionFieldAnonymizer', 'DateFieldAnonymizer', 'CharFieldAnonymizer',
    'DecimalFieldAnonymizer', 'IPAddressFieldAnonymizer', 'StaticValueFieldAnonymizer',
    'MD5TextFieldAnonymizer', 'EmailFieldAnonymizer', 'DeleteFileFieldAnonymizer',
    'ReverseGenericRelationAnonymizer', 'SHA256TextFieldAnonymizer',
    'HashTextFieldAnonymizer', 'GenericRelationAnonymizer', 'IBANFieldAnonymizer', 'SiteIDUsernameFieldAnonymizer',
    'DateTimeFieldAnonymizer', 'IntegerFieldAnonymizer', 'ReplaceFileFieldAnonymizer'
)



import hashlib
import random
import string
import warnings
from typing import (
    Any, Dict, ItemsView, Iterator, KeysView, List, Optional, TYPE_CHECKING, Tuple, Type, Union, ValuesView)

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core import serializers
from django.db import transaction
from django.db.models import Model, QuerySet
from django.db.models.fields import Field

from gdpr.anonymizers.base import BaseAnonymizer, FieldAnonymizer, RelationAnonymizer
from gdpr.fields import Fields
from gdpr.models import AnonymizedData, LegalReason
from gdpr.utils import (
    get_field_or_none, get_reversion_version_model, get_all_parent_objects, get_all_obj_and_parent_versions,
    get_all_obj_and_parent_versions_queryset_list, get_reversion_local_field_dict
)

if TYPE_CHECKING:
    from gdpr.purposes.default import AbstractPurpose

FieldList = Union[List, Tuple, KeysView[str]]  # List, tuple or return of dict keys() method.
FieldMatrix = Union[str, Tuple[Any, ...]]


class ModelAnonymizerMeta(type):
    """
    Metaclass for anonymizers. The main purpose of the metaclass is to register anonymizers and find field anonymizers
    defined in the class as attributes and store it to the fields property.
    """

    def __new__(cls, name, bases, attrs):
        from gdpr.loading import anonymizer_register

        new_obj = super().__new__(cls, name, bases, attrs)

        # Also ensure initialization is only performed for subclasses of ModelAnonymizer
        # (excluding Model class itself).
        parents = [b for b in bases if isinstance(b, ModelAnonymizerMeta)]
        if not parents or not hasattr(new_obj, 'Meta'):
            return new_obj

        fields = getattr(new_obj, 'fields', {})
        anonymizers = getattr(new_obj, 'anonymizers', {})

        for name, obj in attrs.items():
            if isinstance(obj, BaseAnonymizer):
                anonymizers[name] = obj
                if isinstance(obj, FieldAnonymizer):
                    fields[name] = obj

        new_obj.fields = fields
        new_obj.anonymizers = anonymizers

        if not getattr(new_obj.Meta, 'abstract', False):
            anonymizer_register.register(new_obj.Meta.model, new_obj)
            new_obj.Meta.anonymize_reversion = getattr(new_obj.Meta, 'anonymize_reversion', False)
            new_obj.Meta.delete_reversion = getattr(new_obj.Meta, 'delete_reversion', False)
            new_obj.Meta.reversible_anonymization = getattr(new_obj.Meta, 'reversible_anonymization', True)

        return new_obj


class ModelAnonymizerBase(BaseAnonymizer, metaclass=ModelAnonymizerMeta):

    can_anonymize_qs: bool
    fields: Dict[str, FieldAnonymizer]
    anonymizers: Dict[str, BaseAnonymizer]
    _base_encryption_key = None

    class IrreversibleAnonymizerException(Exception):
        pass

    def __init__(self, base_encryption_key: Optional[str] = None):
        self._base_encryption_key = base_encryption_key

    @property
    def model(self) -> Type[Model]:
        return self.Meta.model  # type: ignore

    @property
    def content_type(self) -> ContentType:
        """Get model ContentType"""
        return ContentType.objects.get_for_model(self.model)

    def __getitem__(self, item: str) -> FieldAnonymizer:
        return self.fields[item]

    def __contains__(self, item: str) -> bool:
        return item in self.fields.keys()

    def __iter__(self) -> Iterator[str]:
        for i in self.fields:
            yield i

    def keys(self) -> KeysView[str]:
        return self.fields.keys()

    def items(self) -> ItemsView[str, FieldAnonymizer]:
        return self.fields.items()

    def values(self) -> ValuesView[FieldAnonymizer]:
        return self.fields.values()

    def get(self, *args, **kwargs) -> Union[FieldAnonymizer, Any]:
        return self.fields.get(*args, **kwargs)

    def _get_encryption_key(self, obj, field_name: str):
        """Hash encryption key from `get_encryption_key` and append settings.GDPR_KEY or settings.SECRET_KEY."""
        return hashlib.sha256(
            f'{obj.pk}::{self.get_encryption_key(obj)}::'
            f'{settings.GDPR_KEY if hasattr(settings, "GDPR_KEY") else settings.SECRET_KEY}::{field_name}'.encode(
                'utf-8')).hexdigest()

    def is_reversible(self, obj) -> bool:
        return self.Meta.reversible_anonymization  # type: ignore

    def anonymize_reversion(self, obj, field_names, anonymization: bool):
        from reversion.models import Version

        versions: List[Version] = get_all_obj_and_parent_versions(obj)
        versions_update_data = [
            (
                version,
                {
                    name: self.get_value_from_version(self[name], obj, version, name,
                                                      anonymization=anonymization)
                    for name in field_names
                    if name in get_reversion_local_field_dict(version)
                }
            )
            for version in versions
        ]
        for version, version_dict in versions_update_data:
            self._perform_version_update(version, version_dict)

    def delete_reversion(self, obj, anonymization: bool):
        if anonymization:
            for qs in get_all_obj_and_parent_versions_queryset_list(obj):
                qs.delete()

    def get_encryption_key(self, obj) -> str:
        if not self.is_reversible(obj):
            return ''.join(random.choices(string.digits + string.ascii_letters, k=128))
        if self._base_encryption_key:
            return self._base_encryption_key
        raise NotImplementedError(
            f'The anonymizer \'{self.__class__.__name__}\' does not have `get_encryption_key` method defined or '
            '`base_encryption_key` supplied during anonymization or '
            'reversible_anonymization set to False.')

    def set_base_encryption_key(self, base_encryption_key: str):
        self._base_encryption_key = base_encryption_key

    def is_field_anonymized(self, obj: Model, name: str) -> bool:
        """Check if field have AnonymizedData record"""
        return AnonymizedData.objects.filter(
            field=name, is_active=True, content_type=self.content_type, object_id=str(obj.pk)
        ).exists()

    def get_related_model(self, field_name: str) -> Type[Model]:
        field = get_field_or_none(self.model, field_name)
        if field is None:
            anonymizer = self.anonymizers.get(field_name)
            if anonymizer and isinstance(anonymizer, (ModelAnonymizerBase, RelationAnonymizer)):
                return anonymizer.model
            raise RuntimeError(f'Field \'{field_name}\' is not defined on {str(self.model)}')
        elif hasattr(field, "related_model"):
            return field.related_model
        else:
            raise NotImplementedError(f'Relation {str(field)} not supported yet.')

    def get_value_from_obj(self, field: FieldAnonymizer, obj: Model, name: str, anonymization: bool = True) -> Any:
        return field.get_value_from_obj(obj, name, self._get_encryption_key(obj, name), anonymization=anonymization)

    def get_value_from_version(self, field: FieldAnonymizer, obj: Model, version, name: str,
                               anonymization: bool = True) -> Any:
        return field.get_value_from_version(
            obj, version, name, self._get_encryption_key(obj, name), anonymization=anonymization
        )

    def update_field_as_anonymized(self, obj: Model, name: str, legal_reason: Optional[LegalReason] = None,
                                   anonymization: bool = True):
        if anonymization:
            AnonymizedData.objects.create(object=obj, field=name, expired_reason=legal_reason)
        else:
            AnonymizedData.objects.filter(
                field=name, is_active=True, content_type=self.content_type, object_id=str(obj.pk)
            ).delete()

    def _perform_update(self, obj: Model, updated_data: dict, legal_reason: Optional[LegalReason] = None,
                        anonymization: bool = True):
        for field_name, value in updated_data.items():
            setattr(obj, field_name, value)
        obj.save()
        for field_name in updated_data.keys():
            self.update_field_as_anonymized(obj, field_name, legal_reason, anonymization=anonymization)

    def perform_update(self, obj: Model, updated_data: dict, legal_reason: Optional[LegalReason] = None,
                       anonymization: bool = True):
        self._perform_update(obj, updated_data, legal_reason, anonymization=anonymization)

    @staticmethod
    def _perform_version_update(version, update_data):
        from reversion import revisions

        local_obj = version._object_version.object
        for field, value in update_data.items():
            setattr(local_obj, field, value)
        if hasattr(revisions, '_get_options'):
            version_options = revisions._get_options(get_reversion_version_model(version))
            version_format = version_options.format
            version_fields = version_options.fields
        else:
            version_adapter = revisions.get_adapter(get_reversion_version_model(version))
            version_format = version_adapter.get_serialization_format()
            version_fields = list(version_adapter.get_fields_to_serialize())
        version.serialized_data = serializers.serialize(
            version_format,
            (local_obj,),
            fields=version_fields
        )
        version.save()

    def anonymize_qs(self, qs: QuerySet) -> None:
        raise NotImplementedError()

    def deanonymize_qs(self, qs: QuerySet) -> None:
        raise NotImplementedError()

    def get_related_model_anonymizer_none(self, name):
        anonymizer = self.anonymizers.get(name)

        if anonymizer and isinstance(anonymizer, ModelAnonymizerBase):
            return anonymizer
        elif anonymizer and isinstance(anonymizer, RelationAnonymizer):
            return anonymizer.model_anonymizer
        else:
            return None

    def update_related_anonymizer_fields(self,
                                         field_name: str,
                                         anonymizer: RelationAnonymizer,
                                         obj: Model,
                                         related_fields: Fields,
                                         legal_reason: Optional[LegalReason] = None,
                                         purpose: Optional["AbstractPurpose"] = None,
                                         anonymization: bool = True):
        """
        Anonymize related object defined in related anonymization class.
        Args:
            field_name: name of the model field/property
            anonymizer: relation anonymizer defined it the model anonymization class
            obj: django model instance
            related_fields: fields which will be anonymized
            legal_reason: legal reason which raises anonymization
            purpose: deactivated purpose
            anonymization: anonymize or not
        """
        objs = anonymizer.get_related_objects(obj)
        for related_obj in objs:
            related_fields.anonymizer.update_obj(
                related_obj, legal_reason, purpose, related_fields,
                base_encryption_key=self._get_encryption_key(obj, field_name),
                anonymization=anonymization
            )

    def update_related_model_property_fields(self,
                                             field_name: str,
                                             obj: Model,
                                             related_fields: Fields,
                                             legal_reason: Optional[LegalReason] = None,
                                             purpose: Optional["AbstractPurpose"] = None,
                                             anonymization: bool = True):
        """
        Anonymize related object or objects get from model property.
        Args:
            field_name: name of the model field/property
            obj: django model instance
            related_fields: fields which will be anonymized
            legal_reason: legal reason which raises anonymization
            purpose: deactivated purpose
            anonymization: anonymize or not
        """
        related_attribute = getattr(obj, field_name, None)

        if hasattr(obj.__class__, field_name) and related_attribute is None:
            return
        elif isinstance(related_attribute, Model):
            related_fields.anonymizer.update_obj(
                related_attribute, legal_reason, purpose, related_fields,
                base_encryption_key=self._get_encryption_key(obj, field_name),
                anonymization=anonymization
            )
        elif isinstance(related_attribute, QuerySet):
            for related_obj in related_attribute:
                related_fields.anonymizer.update_obj(
                    related_obj, legal_reason, purpose, related_fields,
                    base_encryption_key=self._get_encryption_key(obj, field_name),
                    anonymization=anonymization
                )
        else:
            warnings.warn(f'Model anonymization discovered unreachable field {field_name} on model'
                          f'{obj.__class__.__name__} on obj {obj} with pk {obj.pk}')

    def update_related_model_fields(self,
                                    field_name: str,
                                    related_metafield: Field,
                                    obj: Model,
                                    related_fields: Fields,
                                    legal_reason: Optional[LegalReason] = None,
                                    purpose: Optional["AbstractPurpose"] = None,
                                    anonymization: bool = True):
        """
        Anonymize related object or objects get from model field.
        Args:
            field_name: name of the model field
            related_metafield: meta infromations get from django model field
            obj: django model instance
            related_fields: fields which will be anonymized
            legal_reason: legal reason which raises anonymization
            purpose: deactivated purpose
            anonymization: anonymize or not
        """
        related_attribute = getattr(obj, field_name, None)
        if related_metafield.one_to_many or related_metafield.many_to_many:
            for related_obj in related_attribute.all():
                related_fields.anonymizer.update_obj(
                    related_obj, legal_reason, purpose, related_fields,
                    base_encryption_key=self._get_encryption_key(obj, field_name),
                    anonymization=anonymization
                )
        elif related_metafield.many_to_one or related_metafield.one_to_one:
            if related_attribute is not None:
                related_fields.anonymizer.update_obj(
                    related_attribute, legal_reason, purpose, related_fields,
                    base_encryption_key=self._get_encryption_key(obj, field_name),
                    anonymization=anonymization
                )
        else:
            warnings.warn(f'Model anonymization discovered unreachable field {field_name} on model'
                          f'{obj.__class__.__name__} on obj {obj} with pk {obj.pk}')

    def update_related_fields(self, parsed_fields: Fields, obj: Model, legal_reason: Optional[LegalReason] = None,
                              purpose: Optional["AbstractPurpose"] = None, anonymization: bool = True):
        for name, related_fields in parsed_fields.related_fields.items():
            related_metafield = get_field_or_none(self.model, name)
            anonymizer = self.anonymizers.get(name)

            if anonymizer and isinstance(anonymizer, RelationAnonymizer):
                self.update_related_anonymizer_fields(
                    name, anonymizer, obj, related_fields, legal_reason, purpose, anonymization
                )
            elif related_metafield:
                self.update_related_model_fields(
                    name, related_metafield, obj, related_fields, legal_reason, purpose, anonymization
                )
            else:
                self.update_related_model_property_fields(
                    name, obj, related_fields, legal_reason, purpose, anonymization
                )

    def update_obj(self, obj: Model, legal_reason: Optional[LegalReason] = None,
                   purpose: Optional["AbstractPurpose"] = None,
                   fields: Union[Fields, FieldMatrix] = '__ALL__',
                   base_encryption_key: Optional[str] = None,
                   anonymization: bool = True):
        if not anonymization and not self.is_reversible(obj):
            raise self.IrreversibleAnonymizerException(f'{self.__class__.__name__} for obj "{obj}" is not reversible.')

        if base_encryption_key:
            self._base_encryption_key = base_encryption_key

        parsed_fields: Fields = Fields(fields, obj.__class__) if not isinstance(fields, Fields) else fields

        if anonymization:
            raw_local_fields = [i for i in parsed_fields.local_fields if not self.is_field_anonymized(obj, i)]
        else:
            raw_local_fields = [i for i in parsed_fields.local_fields if
                                self.is_field_anonymized(obj, i) and self[i].get_is_reversible(obj)]

        if raw_local_fields:
            update_dict = {
                name: self.get_value_from_obj(self[name], obj, name, anonymization) for name in raw_local_fields
            }
            if self.Meta.delete_reversion:  # type: ignore
                self.delete_reversion(obj, anonymization)
            elif self.Meta.anonymize_reversion:  # type: ignore
                self.anonymize_reversion(obj, raw_local_fields, anonymization)
            self.perform_update(obj, update_dict, legal_reason, anonymization=anonymization)

        self.update_related_fields(parsed_fields, obj, legal_reason, purpose, anonymization)

    def anonymize_obj(self, obj: Model, legal_reason: Optional[LegalReason] = None,
                      purpose: Optional["AbstractPurpose"] = None,
                      fields: Union[Fields, FieldMatrix] = '__ALL__', base_encryption_key: Optional[str] = None):

        self.update_obj(obj, legal_reason, purpose, fields, base_encryption_key, anonymization=True)

    def deanonymize_obj(self, obj: Model, fields: Union[Fields, FieldMatrix] = '__ALL__',
                        base_encryption_key: Optional[str] = None):

        self.update_obj(obj, fields=fields, base_encryption_key=base_encryption_key, anonymization=False)


class ModelAnonymizer(ModelAnonymizerBase):
    """
    Default model anonymizer that supports anonymization per object.
    Child must define Meta class with model (like factoryboy)
    """

    can_anonymize_qs = False
    chunk_size = 10000


class DeleteModelAnonymizer(ModelAnonymizer):
    """
    The simpliest anonymization class that is used for removing whole input queryset.

    For anonymization add `__SELF__` to the FieldMatrix.
    """

    can_anonymize_qs = True

    DELETE_FIELD_NAME = '__SELF__'

    def update_obj(self, obj: Model, legal_reason: Optional[LegalReason] = None,
                   purpose: Optional["AbstractPurpose"] = None,
                   fields: Union[Fields, FieldMatrix] = '__ALL__',
                   base_encryption_key: Optional[str] = None,
                   anonymization: bool = True):
        parsed_fields: Fields = Fields(fields, obj.__class__) if not isinstance(fields, Fields) else fields

        if self.DELETE_FIELD_NAME in parsed_fields.local_fields and anonymization is True:
            self.update_related_fields(parsed_fields, obj, legal_reason, purpose, anonymization)

            obj.__class__.objects.filter(pk=obj.pk).delete()

            if self.Meta.delete_reversion:  # type: ignore
                self.delete_reversion(obj, anonymization)

        elif self.DELETE_FIELD_NAME in parsed_fields.local_fields:
            parsed_fields.local_fields = [i for i in parsed_fields.local_fields if i != self.DELETE_FIELD_NAME]
            super().update_obj(obj, legal_reason, purpose, parsed_fields, base_encryption_key, anonymization)
        else:
            super().update_obj(obj, legal_reason, purpose, parsed_fields, base_encryption_key, anonymization)

    def anonymize_qs(self, qs):
        qs.delete()



from django.contrib.contenttypes.models import ContentType
from django.test import TestCase

from faker import Faker
from gdpr.models import LegalReason
from germanium.tools import assert_equal, assert_not_equal, assert_true, assert_raises
from tests.models import Account, Customer, Email, Payment
from tests.purposes import (
    ACCOUNT_AND_PAYMENT_SLUG, ACCOUNT_SLUG, EMAIL_SLUG, EVERYTHING_SLUG, FACEBOOK_SLUG, FIRST_AND_LAST_NAME_SLUG,
    EmailsPurpose, FacebookPurpose, MarketingPurpose
)
from tests.tests.data import (
    ACCOUNT__NUMBER, ACCOUNT__NUMBER2, ACCOUNT__OWNER, ACCOUNT__OWNER2, CUSTOMER__BIRTH_DATE, CUSTOMER__EMAIL,
    CUSTOMER__EMAIL2, CUSTOMER__EMAIL3, CUSTOMER__FACEBOOK_ID, CUSTOMER__FIRST_NAME, CUSTOMER__IP, CUSTOMER__KWARGS,
    CUSTOMER__LAST_NAME, CUSTOMER__PERSONAL_ID, CUSTOMER__PHONE_NUMBER
)
from tests.tests.utils import AnonymizedDataMixin, NotImplementedMixin


class TestLegalReason(AnonymizedDataMixin, NotImplementedMixin, TestCase):

    def setUp(self):
        self.fake = Faker()

    @classmethod
    def setUpTestData(cls):
        cls.customer: Customer = Customer(**CUSTOMER__KWARGS)
        cls.customer.save()

    def test_create_legal_reson_from_slug(self):
        LegalReason.objects.create_consent(FIRST_AND_LAST_NAME_SLUG, self.customer).save()

        assert_true(LegalReason.objects.filter(
            purpose_slug=FIRST_AND_LAST_NAME_SLUG, source_object_id=self.customer.pk,
            source_object_content_type=ContentType.objects.get_for_model(Customer)).exists())

    def test_expirement_legal_reason(self):
        legal = LegalReason.objects.create_consent(FIRST_AND_LAST_NAME_SLUG, self.customer)
        legal.expire()

        anon_customer = Customer.objects.get(pk=self.customer.pk)

        assert_not_equal(anon_customer.first_name, CUSTOMER__FIRST_NAME)
        self.assertAnonymizedDataExists(anon_customer, "first_name")
        assert_not_equal(anon_customer.last_name, CUSTOMER__LAST_NAME)
        self.assertAnonymizedDataExists(anon_customer, "last_name")
        # make sure only data we want were anonymized
        assert_equal(anon_customer.primary_email_address, CUSTOMER__EMAIL)
        self.assertAnonymizedDataNotExists(anon_customer, "primary_email_address")

    def test_renew_legal_reason(self):
        legal = LegalReason.objects.create_consent(FIRST_AND_LAST_NAME_SLUG, self.customer)
        legal.expire()
        legal.renew()

        anon_customer = Customer.objects.get(pk=self.customer.pk)

        # Non reversible anonymization
        assert_not_equal(anon_customer.first_name, CUSTOMER__FIRST_NAME)
        self.assertAnonymizedDataExists(anon_customer, "first_name")
        assert_not_equal(anon_customer.last_name, CUSTOMER__LAST_NAME)
        self.assertAnonymizedDataExists(anon_customer, "last_name")

    def test_expirement_legal_reason_related(self):
        related_email: Email = Email(customer=self.customer, email=CUSTOMER__EMAIL)
        related_email.save()

        related_email2: Email = Email(customer=self.customer, email=CUSTOMER__EMAIL2)
        related_email2.save()

        related_email3: Email = Email(customer=self.customer, email=CUSTOMER__EMAIL3)
        related_email3.save()

        legal = LegalReason.objects.create_consent(EMAIL_SLUG, self.customer)
        legal.expire()

        anon_customer = Customer.objects.get(pk=self.customer.pk)

        assert_equal(anon_customer.primary_email_address, CUSTOMER__EMAIL)
        self.assertAnonymizedDataNotExists(anon_customer, "primary_email_address")

        # make sure only data we want were anonymized
        assert_equal(anon_customer.first_name, CUSTOMER__FIRST_NAME)
        self.assertAnonymizedDataNotExists(anon_customer, "first_name")

        anon_related_email: Email = Email.objects.get(pk=related_email.pk)

        assert_not_equal(anon_related_email.email, CUSTOMER__EMAIL)
        self.assertAnonymizedDataExists(anon_related_email, "email")

        anon_related_email2: Email = Email.objects.get(pk=related_email2.pk)

        assert_not_equal(anon_related_email2.email, CUSTOMER__EMAIL2)
        self.assertAnonymizedDataExists(anon_related_email2, "email")

        anon_related_email3: Email = Email.objects.get(pk=related_email3.pk)

        assert_not_equal(anon_related_email3.email, CUSTOMER__EMAIL3)
        self.assertAnonymizedDataExists(anon_related_email3, "email")

    def test_renew_legal_reason_related(self):
        related_email: Email = Email(customer=self.customer, email=CUSTOMER__EMAIL)
        related_email.save()

        related_email2: Email = Email(customer=self.customer, email=CUSTOMER__EMAIL2)
        related_email2.save()

        related_email3: Email = Email(customer=self.customer, email=CUSTOMER__EMAIL3)
        related_email3.save()

        legal = LegalReason.objects.create_consent(EMAIL_SLUG, self.customer)
        legal.expire()

        anon_legal = LegalReason.objects.get(pk=legal.pk)
        anon_legal.renew()

        anon_customer = Customer.objects.get(pk=self.customer.pk)

        assert_equal(anon_customer.primary_email_address, CUSTOMER__EMAIL)
        self.assertAnonymizedDataNotExists(anon_customer, "primary_email_address")

        # make sure only data we want were anonymized
        assert_equal(anon_customer.first_name, CUSTOMER__FIRST_NAME)
        self.assertAnonymizedDataNotExists(anon_customer, "first_name")

        anon_related_email: Email = Email.objects.get(pk=related_email.pk)

        assert_equal(anon_related_email.email, CUSTOMER__EMAIL)
        self.assertAnonymizedDataNotExists(anon_related_email, "email")

        anon_related_email2: Email = Email.objects.get(pk=related_email2.pk)

        assert_equal(anon_related_email2.email, CUSTOMER__EMAIL2)
        self.assertAnonymizedDataNotExists(anon_related_email2, "email")

        anon_related_email3: Email = Email.objects.get(pk=related_email3.pk)

        assert_equal(anon_related_email3.email, CUSTOMER__EMAIL3)
        self.assertAnonymizedDataNotExists(anon_related_email3, "email")

    def test_expirement_legal_reason_two_level_related(self):
        account_1: Account = Account(customer=self.customer, number=ACCOUNT__NUMBER, owner=ACCOUNT__OWNER)
        account_1.save()
        account_2: Account = Account(customer=self.customer, number=ACCOUNT__NUMBER2, owner=ACCOUNT__OWNER2)
        account_2.save()

        payment_1: Payment = Payment(account=account_1,
                                     value=self.fake.pydecimal(left_digits=8, right_digits=2, positive=True))
        payment_1.save()
        payment_2: Payment = Payment(account=account_1,
                                     value=self.fake.pydecimal(left_digits=8, right_digits=2, positive=True))
        payment_2.save()

        payment_3: Payment = Payment(account=account_2,
                                     value=self.fake.pydecimal(left_digits=8, right_digits=2, positive=True))
        payment_3.save()
        payment_4: Payment = Payment(account=account_2,
                                     value=self.fake.pydecimal(left_digits=8, right_digits=2, positive=True))
        payment_4.save()

        legal = LegalReason.objects.create_consent(ACCOUNT_AND_PAYMENT_SLUG, self.customer)
        legal.expire()

        anon_account_1: Account = Account.objects.get(pk=account_1.pk)

        assert_not_equal(anon_account_1.number, ACCOUNT__NUMBER)
        self.assertAnonymizedDataExists(anon_account_1, "number")
        assert_not_equal(anon_account_1.owner, ACCOUNT__OWNER)
        self.assertAnonymizedDataExists(anon_account_1, "owner")

        anon_account_2: Account = Account.objects.get(pk=account_2.pk)

        assert_not_equal(anon_account_2.number, ACCOUNT__NUMBER2)
        self.assertAnonymizedDataExists(anon_account_2, "number")
        assert_not_equal(anon_account_2.owner, ACCOUNT__OWNER2)
        self.assertAnonymizedDataExists(anon_account_2, "owner")

        for payment in [payment_1, payment_2, payment_3, payment_4]:
            anon_payment: Payment = Payment.objects.get(pk=payment.pk)

            assert_not_equal(anon_payment.value, payment.value)
            self.assertAnonymizedDataExists(anon_payment, "value")
            assert_not_equal(anon_payment.date, payment.date)
            self.assertAnonymizedDataExists(anon_payment, "date")

    def test_renew_legal_reason_two_level_related(self):
        account_1: Account = Account(customer=self.customer, number=ACCOUNT__NUMBER, owner=ACCOUNT__OWNER)
        account_1.save()
        account_2: Account = Account(customer=self.customer, number=ACCOUNT__NUMBER2, owner=ACCOUNT__OWNER2)
        account_2.save()

        payment_1: Payment = Payment(account=account_1,
                                     value=self.fake.pydecimal(left_digits=8, right_digits=2, positive=True))
        payment_1.save()
        payment_2: Payment = Payment(account=account_1,
                                     value=self.fake.pydecimal(left_digits=8, right_digits=2, positive=True))
        payment_2.save()

        payment_3: Payment = Payment(account=account_2,
                                     value=self.fake.pydecimal(left_digits=8, right_digits=2, positive=True))
        payment_3.save()
        payment_4: Payment = Payment(account=account_2,
                                     value=self.fake.pydecimal(left_digits=8, right_digits=2, positive=True))
        payment_4.save()

        legal = LegalReason.objects.create_consent(ACCOUNT_AND_PAYMENT_SLUG, self.customer)
        legal.expire()

        anon_legal = LegalReason.objects.get(pk=legal.pk)
        anon_legal.renew()

        anon_account_1: Account = Account.objects.get(pk=account_1.pk)

        assert_equal(anon_account_1.number, ACCOUNT__NUMBER)
        self.assertAnonymizedDataNotExists(anon_account_1, "number")
        assert_equal(anon_account_1.owner, ACCOUNT__OWNER)
        self.assertAnonymizedDataNotExists(anon_account_1, "owner")

        anon_account_2: Account = Account.objects.get(pk=account_2.pk)

        assert_equal(anon_account_2.number, ACCOUNT__NUMBER2)
        self.assertAnonymizedDataNotExists(anon_account_2, "number")
        assert_equal(anon_account_2.owner, ACCOUNT__OWNER2)
        self.assertAnonymizedDataNotExists(anon_account_2, "owner")

        for payment in [payment_1, payment_2, payment_3, payment_4]:
            anon_payment: Payment = Payment.objects.get(pk=payment.pk)

            assert_equal(anon_payment.value, payment.value)
            self.assertAnonymizedDataNotExists(anon_payment, "value")
            assert_equal(anon_payment.date, payment.date)
            self.assertAnonymizedDataNotExists(anon_payment, "date")

    def test_email_purpose(self):
        LegalReason.objects.create_consent(EMAIL_SLUG, self.customer)

        EmailsPurpose().anonymize_obj(obj=self.customer, fields=("primary_email_address",))

        anon_customer = Customer.objects.get(pk=self.customer.pk)

        assert_equal(anon_customer.primary_email_address, CUSTOMER__EMAIL)
        self.assertAnonymizedDataNotExists(self.customer, 'primary_email_address')

    def test_email_purpose_related(self):
        LegalReason.objects.create_consent(EMAIL_SLUG, self.customer)

        related_email: Email = Email(customer=self.customer, email=CUSTOMER__EMAIL)
        related_email.save()

        EmailsPurpose().anonymize_obj(obj=self.customer, fields=("primary_email_address",))

        anon_customer = Customer.objects.get(pk=self.customer.pk)

        assert_equal(anon_customer.primary_email_address, CUSTOMER__EMAIL)
        self.assertAnonymizedDataNotExists(anon_customer, 'primary_email_address')

        anon_related_email: Email = Email.objects.get(pk=related_email.pk)

        assert_equal(anon_related_email.email, CUSTOMER__EMAIL)
        self.assertAnonymizedDataNotExists(anon_related_email, 'email')

    def test_legal_reason_hardcore(self):
        related_email: Email = Email(customer=self.customer, email=CUSTOMER__EMAIL)
        related_email.save()

        related_email2: Email = Email(customer=self.customer, email=CUSTOMER__EMAIL2)
        related_email2.save()

        account: Account = Account(customer=self.customer, number=ACCOUNT__NUMBER, owner=ACCOUNT__OWNER)
        account.save()

        payment: Payment = Payment(account=account,
                                   value=self.fake.pydecimal(left_digits=8, right_digits=2, positive=True))
        payment.save()

        LegalReason.objects.create_consent(FIRST_AND_LAST_NAME_SLUG, self.customer)
        LegalReason.objects.create_consent(EMAIL_SLUG, self.customer)
        LegalReason.objects.create_consent(ACCOUNT_SLUG, self.customer)
        legal = LegalReason.objects.create_consent(EVERYTHING_SLUG, self.customer)
        legal.expire()

        anon_customer: Customer = Customer.objects.get(pk=self.customer.pk)
        anon_related_email: Email = Email.objects.get(pk=related_email.pk)
        anon_related_email2: Email = Email.objects.get(pk=related_email2.pk)
        anon_account: Account = Account.objects.get(pk=account.pk)
        anon_payment: Payment = Payment.objects.get(pk=payment.pk)

        # Customer - partialy anonymized
        assert_equal(anon_customer.first_name, CUSTOMER__FIRST_NAME)
        self.assertAnonymizedDataNotExists(anon_customer, 'first_name')
        assert_equal(anon_customer.last_name, CUSTOMER__LAST_NAME)
        self.assertAnonymizedDataNotExists(anon_customer, 'last_name')
        assert_not_equal(anon_customer.primary_email_address, CUSTOMER__EMAIL)
        self.assertAnonymizedDataExists(anon_customer, 'primary_email_address')

        assert_not_equal(anon_customer.birth_date, CUSTOMER__BIRTH_DATE)
        self.assertAnonymizedDataExists(anon_customer, 'birth_date')
        assert_not_equal(anon_customer.personal_id, CUSTOMER__PERSONAL_ID)
        self.assertAnonymizedDataExists(anon_customer, 'personal_id')
        assert_not_equal(anon_customer.phone_number, CUSTOMER__PHONE_NUMBER)
        self.assertAnonymizedDataExists(anon_customer, 'phone_number')
        assert_not_equal(anon_customer.facebook_id, CUSTOMER__FACEBOOK_ID)
        self.assertAnonymizedDataExists(anon_customer, 'facebook_id')
        assert_not_equal(anon_customer.last_login_ip, CUSTOMER__IP)
        self.assertAnonymizedDataExists(anon_customer, 'last_login_ip')

        # Email - not anonymized
        assert_equal(anon_related_email.email, CUSTOMER__EMAIL)
        self.assertAnonymizedDataNotExists(anon_related_email, 'email')
        assert_equal(anon_related_email2.email, CUSTOMER__EMAIL2)
        self.assertAnonymizedDataNotExists(anon_related_email2, 'email')

        # Account - not anonymized
        assert_equal(anon_account.number, ACCOUNT__NUMBER)
        self.assertAnonymizedDataNotExists(anon_account, "number")
        assert_equal(anon_account.owner, ACCOUNT__OWNER)
        self.assertAnonymizedDataNotExists(anon_account, "owner")

        # Payment - fully anonymized
        assert_not_equal(anon_payment.value, payment.value)
        self.assertAnonymizedDataExists(anon_payment, "value")
        assert_not_equal(anon_payment.date, payment.date)
        self.assertAnonymizedDataExists(anon_payment, "date")

    def test_purpose_source_model_class_should_be_set_with_string(self):
        assert_equal(FacebookPurpose.source_model_class, Customer)

    def test_purpose_source_model_class_should_be_set_directly(self):
        assert_equal(MarketingPurpose.source_model_class, Customer)

    def test_purpose_should_be_set_only_to_the_right_source_model(self):
        LegalReason.objects.create_consent(FACEBOOK_SLUG, self.customer)
        with assert_raises(KeyError):
            LegalReason.objects.create_consent(FACEBOOK_SLUG, Email(customer=self.customer, email=CUSTOMER__EMAIL))

    def test_facebook_purpose_should_anonymize_customer_with_facebook_id(self):
        customer = Customer.objects.get(pk=self.customer.pk)
        legal_reason = LegalReason.objects.create_consent(FACEBOOK_SLUG, customer)
        legal_reason.save()
        legal_reason.expire()
        customer.refresh_from_db()
        assert_not_equal(customer.first_name, CUSTOMER__FIRST_NAME)

    def test_facebook_purpose_should_anonymize_customer_without_facebook_id(self):
        customer = Customer.objects.get(pk=self.customer.pk)
        legal_reason = LegalReason.objects.create_consent(FACEBOOK_SLUG, customer)
        customer.facebook_id = None
        customer.save()
        legal_reason.save()
        legal_reason.expire()
        customer.refresh_from_db()
        assert_equal(customer.first_name, CUSTOMER__FIRST_NAME)



import json
from datetime import timedelta
from typing import Any, Callable, Optional, Union

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.core.files.base import ContentFile, File
from django.db.models.fields.files import FieldFile
from django.utils.inspect import func_supports_parameter
from unidecode import unidecode

from gdpr.anonymizers.base import FieldAnonymizer, NumericFieldAnonymizer
from gdpr.encryption import (
    JSON_SAFE_CHARS, decrypt_email_address, decrypt_text, encrypt_email_address, encrypt_text, numerize_key,
    translate_iban, translate_number, translate_text)
from gdpr.ipcypher import decrypt_ip, encrypt_ip
from gdpr.utils import get_number_guess_len


class FunctionFieldAnonymizer(FieldAnonymizer):
    """
    Use this field anonymization for defining in place lambda anonymization method.

    Example:
    ```
    secret_code = FunctionFieldAnonymizer(lambda self, x, key: x**2)
    ```

    If you want to supply anonymization and deanonymization you can do following:
    ```
    secret_code = FunctionFieldAnonymizer(
        func=lambda self, x, key: x**2+self.get_numeric_encryption_key(key),
        deanonymize_func=lambda a, x, key: x**2-a.get_numeric_encryption_key(key)
        )
    ```
    """

    anon_func: Callable
    deanonymize_func: Optional[Callable] = None
    max_anonymization_range: int

    def __init__(self,
                 anon_func: Union[Callable[[Any, Any], Any], Callable[["FunctionFieldAnonymizer", Any, Any], Any]],
                 deanonymize_func: Callable[["FunctionFieldAnonymizer", Any, Any], Any] = None,
                 *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if callable(anon_func):
            self.anon_func = anon_func  # type: ignore
        else:
            raise ImproperlyConfigured('Supplied func is not callable.')

        if callable(deanonymize_func):
            self.deanonymize_func = deanonymize_func
        elif deanonymize_func is not None:
            raise ImproperlyConfigured('Supplied deanonymize_func is not callable.')

    def get_numeric_encryption_key(self, encryption_key: str) -> int:
        return numerize_key(encryption_key) % self.max_anonymization_range

    def get_encrypted_value(self, value, encryption_key: str):
        if self.deanonymize_func is None:
            return self.anon_func(value, encryption_key)
        else:
            return self.anon_func(self, value, encryption_key)

    def get_is_reversible(self, obj=None, raise_exception: bool = False):
        is_reversible = self.deanonymize_func is not None
        if not is_reversible:
            raise self.IrreversibleAnonymizationException
        return is_reversible

    def get_decrypted_value(self, value, encryption_key: str):
        if not self.get_is_reversible():
            raise self.IrreversibleAnonymizationException()
        else:
            return self.deanonymize_func(self, value, encryption_key)  # type: ignore


class DateTimeFieldAnonymizer(NumericFieldAnonymizer):
    """
    Anonymization for DateTimeField.

    """

    max_anonymization_range = 365 * 24 * 60 * 60

    def get_encrypted_value(self, value, encryption_key: str):
        return value - timedelta(seconds=(self.get_numeric_encryption_key(encryption_key) + 1))

    def get_decrypted_value(self, value, encryption_key: str):
        return value + timedelta(seconds=(self.get_numeric_encryption_key(encryption_key) + 1))


class DateFieldAnonymizer(NumericFieldAnonymizer):
    """
    Anonymization for DateField.

    """

    max_anonymization_range = 365

    def get_encrypted_value(self, value, encryption_key: str):
        return value - timedelta(days=(self.get_numeric_encryption_key(encryption_key) + 1))

    def get_decrypted_value(self, value, encryption_key: str):
        return value + timedelta(days=(self.get_numeric_encryption_key(encryption_key) + 1))


class CharFieldAnonymizer(FieldAnonymizer):
    """
    Anonymization for CharField.

    transliterate - The CharFieldAnonymizer encrypts only ASCII chars and non-ascii chars are left the same e.g.:
    `François` -> `rbTTç]3d` if True the original text is transliterated e.g. `François` -> 'Francois' -> `rbTTQ9Zg`.
    """

    transliterate = False
    empty_values = [None, '']

    def __init__(self, *args, transliterate: bool = False, **kwargs):
        self.transliterate = transliterate
        super().__init__(*args, **kwargs)

    def get_encrypted_value(self, value, encryption_key: str):
        return encrypt_text(encryption_key, value if not self.transliterate else unidecode(value))

    def get_decrypted_value(self, value, encryption_key: str):
        return decrypt_text(encryption_key, value)


class EmailFieldAnonymizer(FieldAnonymizer):

    empty_values = [None, '']

    def get_encrypted_value(self, value, encryption_key: str):
        return encrypt_email_address(encryption_key, value)

    def get_decrypted_value(self, value, encryption_key: str):
        return decrypt_email_address(encryption_key, value)


class DecimalFieldAnonymizer(NumericFieldAnonymizer):
    """
    Anonymization for CharField.
    """

    def get_encrypted_value(self, value, encryption_key: str):
        return translate_number(str(self.get_numeric_encryption_key(encryption_key)), value)

    def get_decrypted_value(self, value, encryption_key: str):
        return translate_number(str(self.get_numeric_encryption_key(encryption_key)), value, encrypt=False)


class IntegerFieldAnonymizer(NumericFieldAnonymizer):
    """
    Anonymization for IntegerField.
    """

    def get_encrypted_value(self, value, encryption_key: str):
        return translate_number(str(self.get_numeric_encryption_key(encryption_key)), value)

    def get_decrypted_value(self, value, encryption_key: str):
        return translate_number(str(self.get_numeric_encryption_key(encryption_key)), value, encrypt=False)


class IPAddressFieldAnonymizer(FieldAnonymizer):
    """
    Anonymization for GenericIPAddressField.

    Works for both ipv4 and ipv6.
    """

    empty_values = [None, '']

    def get_encrypted_value(self, value, encryption_key: str):
        return encrypt_ip(encryption_key, value)

    def get_decrypted_value(self, value, encryption_key: str):
        return decrypt_ip(encryption_key, value)


class IBANFieldAnonymizer(FieldAnonymizer):
    """
    Field anonymizer for International Bank Account Number.
    """

    empty_values = [None, '']

    def get_decrypted_value(self, value: Any, encryption_key: str):
        return translate_iban(encryption_key, value)

    def get_encrypted_value(self, value: Any, encryption_key: str):
        return translate_iban(encryption_key, value, False)


class JSONFieldAnonymizer(FieldAnonymizer):
    """
    Anonymization for JSONField.
    """

    empty_values = [None, '']

    def get_numeric_encryption_key(self, encryption_key: str, value: Union[int, float] = None) -> int:
        if value is None:
            return numerize_key(encryption_key)
        return numerize_key(encryption_key) % 10 ** get_number_guess_len(value)

    def anonymize_json_value(self, value: Union[list, dict, bool, None, str, int, float],
                             encryption_key: str,
                             anonymize: bool = True) -> Union[list, dict, bool, None, str, int, float]:
        if value is None:
            return None
        elif type(value) is str:
            return translate_text(encryption_key, value, anonymize, JSON_SAFE_CHARS)  # type: ignore
        elif type(value) is int:
            return translate_number(encryption_key, value, anonymize)  # type: ignore
        elif type(value) is float:
            # We cannot safely anonymize floats
            return value
        elif type(value) is dict:
            return {key: self.anonymize_json_value(item, encryption_key, anonymize) for key, item in
                    value.items()}  # type: ignore
        elif type(value) is list:
            return [self.anonymize_json_value(item, encryption_key, anonymize) for item in value]  # type: ignore
        elif type(value) is bool and self.get_numeric_encryption_key(encryption_key) % 2 == 0:
            return not value
        return value

    def get_encrypted_value(self, value, encryption_key: str):
        if type(value) not in [dict, list, str]:
            raise ValidationError("JSONFieldAnonymizer encountered unknown type of json. "
                                  "Only python dict and list are supported.")
        if type(value) == str:
            return json.dumps(self.anonymize_json_value(json.loads(value), encryption_key))
        return self.anonymize_json_value(value, encryption_key)

    def get_decrypted_value(self, value, encryption_key: str):
        if type(value) not in [dict, list, str]:
            raise ValidationError("JSONFieldAnonymizer encountered unknown type of json. "
                                  "Only python dict and list are supported.")
        if type(value) == str:
            return json.dumps(self.anonymize_json_value(json.loads(value), encryption_key, anonymize=False))
        return self.anonymize_json_value(value, encryption_key, anonymize=False)


class StaticValueFieldAnonymizer(FieldAnonymizer):
    """
    Static value anonymizer replaces value with defined static value.
    """

    is_reversible = False
    empty_values = [None, '']

    def __init__(self, value: Any, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.value: Any = value

    def get_encrypted_value(self, value: Any, encryption_key: str) -> Any:
        return self.value


class SiteIDUsernameFieldAnonymizer(FieldAnonymizer):
    """
    Encrypts username in format 1:foo@bar.com
    """

    empty_values = [None, '']

    def get_encrypted_value(self, value, encryption_key: str):
        split = value.split(':', 1)
        if len(split) == 2:
            return f'{split[0]}:{encrypt_email_address(encryption_key, split[1])}'
        return encrypt_email_address(encryption_key, value)

    def get_decrypted_value(self, value, encryption_key: str):
        split = value.split(':', 1)
        if len(split) == 2:
            return f'{split[0]}:{decrypt_email_address(encryption_key, split[1])}'
        return decrypt_email_address(encryption_key, value)


class FileFieldAnonymizer(FieldAnonymizer):
    """
    Base class for all FileFieldAnonymizers.

    Overrides ``get_is_value_empty`` to check for files.
    """

    def get_is_value_empty(self, value):
        return self.get_ignore_empty_values(value) and not bool(value)


class DeleteFileFieldAnonymizer(FileFieldAnonymizer):
    """
    One way anonymization of FileField.
    """

    is_reversible = False

    def get_encrypted_value(self, value: Any, encryption_key: str):
        value.delete(save=False)
        return value


class ReplaceFileFieldAnonymizer(FileFieldAnonymizer):
    """
    One way anonymization of FileField.
    """

    is_reversible = False
    replacement_file: Optional[str] = None

    def __init__(self, replacement_file: Optional[str] = None, *args, **kwargs):
        if replacement_file is not None:
            self.replacement_file = replacement_file
        super().__init__(*args, **kwargs)

    def get_replacement_file(self):
        if self.replacement_file is not None:
            return File(open(self.replacement_file, "rb"))
        elif getattr(settings, "GDPR_REPLACE_FILE_PATH", None) is not None:
            return File(open(getattr(settings, "GDPR_REPLACE_FILE_PATH"), "rb"))
        else:
            return ContentFile("THIS FILE HAS BEEN ANONYMIZED.")

    def get_encrypted_value(self, value: FieldFile, encryption_key: str):
        file_name = value.name
        value.delete(save=False)
        file = self.get_replacement_file()

        if func_supports_parameter(value.storage.save, 'max_length'):
            value.name = value.storage.save(file_name, file, max_length=value.field.max_length)
        else:
            #  Backwards compatibility removed in Django 1.10
            value.name = value.storage.save(file_name, file)
        setattr(value.instance, value.field.name, value.name)

        value._size = file.size  # Django 1.8 + 1.9
        value._committed = True
        file.close()

        return value



"""
Models for test app:

Customer
  - Email
  - Address
  - AccountNumber
    - Payment

"""
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.translation import gettext_lazy as _

from gdpr.mixins import AnonymizationModel
from gdpr.utils import is_reversion_installed
from tests.validators import CZBirthNumberValidator, BankAccountValidator


class CustomerRegistration(AnonymizationModel):
    email_address = models.EmailField(blank=True, null=True)


class Customer(AnonymizationModel):
    # Keys for pseudoanonymization
    first_name = models.CharField(max_length=256)
    last_name = models.CharField(max_length=256)
    primary_email_address = models.EmailField(blank=True, null=True)

    full_name = models.CharField(max_length=256, blank=True, null=True)
    birth_date = models.DateField(blank=True, null=True)
    personal_id = models.CharField(max_length=10, blank=True, null=True, validators=[CZBirthNumberValidator])
    phone_number = models.CharField(max_length=9, blank=True, null=True)
    facebook_id = models.CharField(
        max_length=256, blank=True, null=True,
        verbose_name=_("Facebook ID"), help_text=_("Facebook ID used for login via Facebook."))
    last_login_ip = models.GenericIPAddressField(blank=True, null=True)

    @property
    def other_registrations(self):
        return CustomerRegistration.objects.filter(email_address=self.primary_email_address).order_by('-pk')[1:]

    @property
    def last_registration(self):
        return CustomerRegistration.objects.filter(email_address=self.primary_email_address).order_by('pk').last()

    def save(self, *args, **kwargs):
        """Just helper method for saving full name.

        You can ignore this method.
        """
        self.full_name = "%s %s" % (self.first_name, self.last_name)
        super().save(*args, **kwargs)

    def __str__(self):
        return f"{self.first_name} {self.last_name}"


class Email(AnonymizationModel):
    """Example on anonymization on related field."""
    customer = models.ForeignKey(Customer, on_delete=models.CASCADE, related_name="emails")
    email = models.EmailField(blank=True, null=True)


class Address(AnonymizationModel):
    customer = models.ForeignKey(Customer, on_delete=models.CASCADE, related_name="addresses")
    street = models.CharField(max_length=256, blank=True, null=True)
    house_number = models.CharField(max_length=20, blank=True, null=True)
    city = models.CharField(max_length=256, blank=True, null=True)
    post_code = models.CharField(max_length=6, blank=True, null=True)


class Account(AnonymizationModel):
    customer = models.ForeignKey(Customer, on_delete=models.CASCADE, related_name="accounts")
    number = models.CharField(max_length=256, blank=True, null=True, validators=[BankAccountValidator])
    IBAN = models.CharField(max_length=34, blank=True, null=True)
    swift = models.CharField(max_length=11, blank=True, null=True)
    owner = models.CharField(max_length=256, blank=True, null=True)


class Payment(AnonymizationModel):
    """Down the rabbit hole multilevel relations."""
    account = models.ForeignKey(Account, on_delete=models.CASCADE, related_name="payments")
    value = models.DecimalField(blank=True, null=True, decimal_places=2, max_digits=10)
    date = models.DateField(auto_now_add=True)


class ContactForm(AnonymizationModel):
    email = models.EmailField()
    full_name = models.CharField(max_length=256)


class Note(AnonymizationModel):
    note = models.TextField()
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey('content_type', 'object_id')


class Avatar(AnonymizationModel):
    customer = models.ForeignKey(Customer, on_delete=models.CASCADE, related_name="avatars")
    image = models.FileField()


class TopParentA(AnonymizationModel):
    name = models.CharField(max_length=250)


class ParentB(TopParentA):
    birth_date = models.DateField()


class ParentC(ParentB):
    first_name = models.CharField(max_length=250)


class ExtraParentD(AnonymizationModel):
    id_d = models.AutoField(primary_key=True, editable=False)
    note = models.CharField(max_length=250)


class ChildE(ParentC, ExtraParentD):
    last_name = models.CharField(max_length=250)


if is_reversion_installed():
    from reversion import revisions as reversion

    reversion.register(Customer)
    reversion.register(Email)
    reversion.register(Address)
    reversion.register(Account)
    reversion.register(Payment)
    reversion.register(ContactForm)
    reversion.register(Note)
    reversion.register(TopParentA)
    reversion.register(ParentB, follow=('topparenta_ptr',))
    reversion.register(ParentC, follow=('parentb_ptr',))
    reversion.register(ExtraParentD)
    reversion.register(ChildE, follow=('parentc_ptr', 'extraparentd_ptr'))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.26 on 2020-02-17 13:26
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('gdpr', '0008_migration'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='legalreason',
            name='is_active',
        ),
    ]
